from datetime import date, datetime, timedelta, timezone
from typing import Union

import dateutil.parser
from dateutil.tz import tzutc

from preheat_open import TIMEZONE

utc = timezone.utc


def timestep_start(step, t):

    if step in ["second", "1s"]:
        t_start = t.replace(microsecond=0)
    elif step == "15s":
        sec_start = int(t.second / 15) * 15
        t_start = t.replace(microsecond=0, second=sec_start)
    elif step == "30s":
        sec_start = int(t.second / 30) * 30
        t_start = t.replace(microsecond=0, second=sec_start)
    elif step in ["minute", "1min"]:
        t_start = t.replace(microsecond=0, second=0)
    elif step == "5min":
        min_start = int(t.minute / 5) * 5
        t_start = t.replace(microsecond=0, second=0, minute=min_start)
    elif step == "15min":
        min_start = int(t.minute / 15) * 15
        t_start = t.replace(microsecond=0, second=0, minute=min_start)
    elif step == "30min":
        min_start = int(t.minute / 30) * 30
        t_start = t.replace(microsecond=0, second=0, minute=min_start)
    elif step == "hour":
        t_start = t.replace(microsecond=0, second=0, minute=0)
    elif step == "day":
        t_start = t.replace(microsecond=0, second=0, minute=0, hour=0)
    elif step == "month":
        t_start = t.replace(microsecond=0, second=0, minute=0, hour=0, day=1)
    elif step == "year":
        t_start = t.replace(microsecond=0, second=0, minute=0, hour=0, day=1, month=1)
    else:
        raise Exception("Unknown step: " + step)

    return t_start


def now(step=None, tz=TIMEZONE):
    t = datetime.now(tz=tz)
    if step is None:
        return t
    else:
        return timestep_start(step, t)


def __enforce_imports():
    date.today() + timedelta(days=2)


def datetime_convert(param):
    if isinstance(param, datetime):
        dt = param
    elif isinstance(param, str):
        dt = dateutil.parser.parse(param)
    else:
        raise TypeError(f"No conversion from type: {type(param)}")

    return dt if dt.tzinfo is not None else dt.astimezone(TIMEZONE)


def sanitise_datetime_input(t: Union[datetime, str]) -> datetime:

    if isinstance(t, str):
        out = datetime_convert(t)
    else:
        out = t
    return out.astimezone(tzutc())


def time_resolution_aliases(time_resolution):
    if time_resolution in ["minute", "5min"]:
        return "5T"
    elif time_resolution == "hour":
        return "H"
    elif time_resolution == "day":
        return "D"
    elif time_resolution == "week":
        return "W"
    elif time_resolution == "month":
        return "MS"
    elif time_resolution == "year":
        return "YS"
    else:
        return None


def convenience_result_list_shortener(result):
    n_results = len(result)
    if n_results > 1:
        return result
    elif n_results == 0:
        return None
    else:
        return result[0]


def list_to_string(list2use, separator=",") -> str:
    """
    Helper function to turn list into string, e.g. comma separated (default).
    """

    if isinstance(list2use, list):
        res = separator.join(map(str, list2use))
    else:
        raise TypeError("Input list2use must be a list")
    return res



import math
from functools import reduce

import numpy as np
import pandas as pd

from .api import (
    DATETIME_FORMAT,
    MAX_POINTS_PER_REQ,
    TIMEZONE,
    APIDataExtractError,
    api_get,
)
from .helpers import list_to_string, sanitise_datetime_input


def load_box_data(
    component_map, start_date, end_date, time_resolution="minute"
) -> pd.DataFrame:
    # Parse strings to datetime object

    start_date = sanitise_datetime_input(start_date)
    end_date = sanitise_datetime_input(end_date)
    __check_start_end_date(start_date, end_date)

    path = "measurements"
    cid_key = "cids"
    cids = list(component_map.keys())

    if not cids:
        return pd.DataFrame()

    resp = __perform_requests(
        path, cid_key, cids, start_date, end_date, time_resolution, "csv"
    )

    return __extract_data(resp, "cid", component_map, "csv")


def load_model_data(
    component_map, start_date, end_date, time_resolution="raw"
) -> pd.DataFrame:
    # Parse strings to datetime object
    start_date = sanitise_datetime_input(start_date)
    end_date = sanitise_datetime_input(end_date)
    __check_start_end_date(start_date, end_date)

    path = "units/measurements"
    id_key = "ids"
    ids = list(component_map.keys())

    if not ids or len(ids) == 0:
        return pd.DataFrame()

    if time_resolution != "raw":
        # Do chunked requests
        resp = __perform_requests(
            path, id_key, ids, start_date, end_date, time_resolution, "csv"
        )
    else:
        # Do single request
        resp = __make_request(
            path, {id_key: ids}, start_date, end_date, time_resolution, "csv"
        )

    return __extract_data(resp, "id", component_map, "csv")


def load_weather_data(
    location_id,
    component_map,
    start_date,
    end_date,
    time_resolution="hour",
    components=None,
) -> pd.DataFrame:
    # Parse strings to datetime object
    start_date = sanitise_datetime_input(start_date)
    end_date = sanitise_datetime_input(end_date)
    __check_start_end_date(start_date, end_date)

    path = f"weather/{location_id}"
    cid_key = "type_ids"
    cids = list(component_map.keys())

    # If minutes, then we need to resample
    if time_resolution in ["minute", "5min", "raw"]:
        time_resolution = "hour"
        resample = True
    else:
        resample = False

    resp = __perform_requests(
        path, cid_key, cids, start_date, end_date, time_resolution, "csv"
    )

    if resample:
        return (
            __extract_data(resp, "type_id", component_map, "csv")
            .resample("5min")
            .ffill()
        )

    return __extract_data(resp, "type_id", component_map, "csv")


def perform_requests(
    path, cid_key, cids, start_date, end_date, time_resolution, format="csv"
):
    return __perform_requests(
        path, cid_key, cids, start_date, end_date, time_resolution, format
    )


def __perform_requests(
    path, cid_key, cids, start_date, end_date, time_resolution, format="csv"
):
    # Split CIDs into chunks
    # NOTE: we reduce number of CIDs pr. REQ to confine w. MAX_POINTS_PER_REQ
    # This sets the limit to 90000 points on a single CID; which for 5min.
    # is roughly 300 days of data.
    time_lookup = {
        "raw": 10,
        "minute": 300,
        "5min": 300,
        "hour": 3600,
        "day": 86400,
        "week": 604800,
        "month": 16934400,
        "year": 6181056000,
    }

    end_date_in = sanitise_datetime_input(end_date)
    start_date_in = sanitise_datetime_input(start_date)

    dt = end_date_in - start_date_in

    N_points_requested = len(cids) * dt.total_seconds() / time_lookup[time_resolution]

    if N_points_requested < 1:
        reqs = []
    else:
        cid_chunk_size = (MAX_POINTS_PER_REQ / N_points_requested) * len(cids)

        cid_chunks = [
            cids[x : x + max(int(cid_chunk_size), 1)]
            for x in range(0, len(cids), max(int(cid_chunk_size), 1))
        ]
        time_chunks = math.ceil(1 / cid_chunk_size)
        dt_chunk = (end_date_in - start_date_in) / np.fmax(1, time_chunks)

        reqs = []
        for cid_chunk in cid_chunks:
            for i in range(time_chunks):
                reqs.append(
                    __make_request(
                        path,
                        {cid_key: list_to_string(cid_chunk)},
                        start_date + i * dt_chunk,
                        start_date + (i + 1) * dt_chunk,
                        time_resolution,
                        format,
                    )
                )

    return reqs


def __make_request(path, payload, start_date, end_date, time_resolution, format="csv"):
    payload["start_time"] = start_date.isoformat()
    payload["end_time"] = end_date.isoformat()

    if time_resolution == "5min":
        time_resolution = "minute"

    payload["time_resolution"] = time_resolution

    return api_get(path, payload=payload, out=format)


def extract_data(resp, id_key, component_map, format="csv"):
    return __extract_data(resp, id_key, component_map, format)


def __extract_data(resp, id_key, component_map, format="csv"):

    kwargs_to_datetime = dict(utc=True, format=DATETIME_FORMAT)
    # If this is having an issue, use "infer_datetime_format=True" instead of "format=..."

    if len(resp) < 1:
        # Short circuit all in case of empty response
        return pd.DataFrame()
    elif format == "json":
        if isinstance(resp, list):
            # Initialize empty response that we can append to
            resp_new = {cid: [] for cid in component_map.keys()}
            for req in resp:
                for cid, values in req.items():
                    resp_new[cid] += values
            resp = resp_new

        if "message" in resp:
            raise APIDataExtractError(resp["message"])

        dfs = []
        for cid, name in component_map.items():
            time = pd.to_datetime(
                [data["time"] for data in resp[cid]], **kwargs_to_datetime
            )
            value = [data["value"] for data in resp[cid]]

            df_i = pd.DataFrame(value, columns=[name], index=time)
            dfs.append(df_i)

    elif format == "csv":
        # Join response into one
        if isinstance(resp, list):
            resp_new = None
            for df in resp:
                resp_new = df if resp_new is None else resp_new.append(df)
            resp = resp_new

        # Split
        dfs = []
        for cid, name in component_map.items():
            df_i = resp[(resp[id_key] == int(cid))].copy()
            df_i.loc[:, "time"] = pd.to_datetime(df_i["time"], **kwargs_to_datetime)
            df_i = df_i.set_index("time")
            df_i = df_i.filter(items=["value"])
            df_i = df_i.rename(columns={"value": name})
            df_i.index = df_i.index.rename(None)
            dfs.append(df_i)

    elif format == "dataframe":
        # Split
        dfs = []
        for cid, name in component_map.items():
            df_i = resp[(resp[id_key] == int(cid))].copy()
            df_i.loc[:, "time"] = pd.to_datetime(df_i["time"], **kwargs_to_datetime)
            df_i = df_i.set_index("time")
            df_i = df_i.filter(items=["value"])
            df_i = df_i.rename(columns={"value": name})
            df_i.index = df_i.index.rename(None)
            dfs.append(df_i)

    else:
        raise TypeError("Unsupported format")

    df = reduce(
        lambda x, y: pd.merge(x, y, left_index=True, right_index=True, how="outer"), dfs
    )
    return df


def __check_start_end_date(start, end):
    if end < start:
        raise ValueError("End date must be AFTER start date")



# Imports for backwards compatibility of dependent code
# noinspection PyUnresolvedReferences
import os
import re
import sys

from . import _version
from .api import (
    BASE_URL,
    MAX_CIDS_PER_REQ,
    MAX_POINTS_PER_REQ,
    TIMEZONE,
    AccessDeniedError,
    APIDataExtractError,
    APIKeyMissingError,
    load_configuration,
    set_api_key,
)

# noinspection PyUnresolvedReferences
from .building import Building, available_buildings

# noinspection PyUnresolvedReferences
from .logging import set_logging_level

# noinspection PyUnresolvedReferences
from .unit import Unit

__version__ = _version.get_versions()["version"]

config = load_configuration()


def running_in_test_mode() -> bool:
    # A function to detect whether or not the system is running with Pytest
    if ("pytest" in sys.modules) or ("PYTEST_CURRENT_TEST" in os.environ):
        return True
    elif any(re.findall(r"pytest|py.test", sys.argv[0])):
        return True
    else:
        return False



import inspect
import logging
import sys
import warnings
from typing import Optional, Union

from .singleton import Singleton

LOG_FORMAT = "%(asctime)-23s  %(levelname)-8s  %(name)-32s  %(message)-160s  .(%(filename)s:%(lineno)d)"


class Logging(metaclass=Singleton):
    def __init__(self, stack_level: int = 0):
        self.stack_level = 2 + stack_level
        if sys.gettrace() is None:
            logging.basicConfig(level=logging.INFO, format=LOG_FORMAT)
        else:
            logging.basicConfig(level=logging.DEBUG, format=LOG_FORMAT)

    def _get_logger(self, failsafe: bool = True, stack_level: int = 0):
        try:
            frame = inspect.stack()[self.stack_level + stack_level]
            mod = inspect.getmodule(frame[0])
            # return logging.getLogger(mod.__package__)
            return logging.getLogger(mod.__name__.split(".")[-1])
        except Exception as e:
            if failsafe is True:
                warnings.warn(
                    "There was an issue in accessing the desired logger - defaulting to a failsafe one"
                )
                return logging.getLogger("Failsafe-log (preheat_open)")
            else:
                raise e

    def set_level(self, level: Union[str, int]) -> None:
        logging.basicConfig(level=level, force=True)

    def error(
        self, msg: str, exception: Optional[Exception] = None, *args, **kwargs
    ) -> None:
        if exception is None:
            for line in msg.split("\n"):
                self._get_logger().error(
                    line, stacklevel=self.stack_level, *args, **kwargs
                )
        else:
            try:
                raise exception
            except Exception as e:
                for line in msg.split("\n"):
                    self._get_logger().exception(
                        line, stacklevel=self.stack_level, *args, **kwargs
                    )

    def warning(self, msg: Union[str, Warning], *args, **kwargs) -> None:
        warnings.warn(msg)
        if isinstance(msg, Warning):
            msg = str(msg)
        for line in msg.split("\n"):
            self._get_logger().warning(
                line, stacklevel=self.stack_level, *args, **kwargs
            )

    def info(self, msg: str, *args, **kwargs) -> None:
        for line in msg.split("\n"):
            self._get_logger().info(line, stacklevel=self.stack_level, *args, **kwargs)

    def debug(self, msg: str, *args, **kwargs) -> None:
        for line in msg.split("\n"):
            self._get_logger().debug(line, stacklevel=self.stack_level, *args, **kwargs)


def logging_level(level: str) -> int:
    """
    Converts a string to a logging level

    :param level: logging level (debug, info, warning, error, critical)
    :type level:
    :return: logging level identifier (in logging package)
    :rtype:
    """
    if isinstance(level, str):
        str_level = level.lower()
        if str_level == "debug":
            log_level = logging.DEBUG
        elif str_level == "info":
            log_level = logging.INFO
        elif str_level == "warning":
            log_level = logging.WARNING
        elif str_level == "error":
            log_level = logging.ERROR
        elif str_level == "critical":
            log_level = logging.CRITICAL
        else:
            raise Exception(f"Illegal logging level ({level})")

        return log_level

    else:
        raise Exception("Only logging levels in string format are supported for now")


def set_logging_level(level: str) -> None:
    """
    Sets the logging level

    :param level: logging level (debug, info, warning, error, critical)
    :type level:
    """
    Logging().set_level(logging_level(level))



import networkx as nx

from .unit import Unit


def generate_unit_graph(building):
    """
    Function to build unit graph from a Building object.
    Iterates through units (nodes) and builds graph based on relations (sub-units).
    Returns a networkx directed multigraph.
    """

    # Empty graph
    G = nx.MultiDiGraph()
    # Number of nodes, initially
    pre = 0

    # First run (top-level units)
    for unit_type, units in building.units.items():
        G.add_nodes_from(
            [(unit.id, {"unit": unit, "name": unit.name}) for unit in units]
        )

    # Note: discovery of sub-units is not optimized (but robust)

    # While new units are discovered
    while pre < G.number_of_nodes():
        # Update number of nodes
        pre = G.number_of_nodes()

        # For each old node (copy of graph)
        for node, data in nx.create_empty_copy(G).nodes.data():
            if hasattr(data["unit"], "sub_units"):
                # Add sub-units as new nodes
                new_nodes = [
                    (unit.id, {"unit": unit, "name": unit.name})
                    for unit in data["unit"].sub_units
                ]
                G.add_nodes_from(new_nodes)

                # Make corresponding edges between nodes
                new_edges = [
                    (node, unit.id, {"weight": 0})
                    for unit in data["unit"].sub_units
                    if not G.has_edge(node, unit.id)
                ]
                G.add_edges_from(new_edges)

    if building.units["main"]:
        # Add edges to Main (heating and hot water)
        G.add_edges_from(
            [
                (building.units["main"][0].id, node, {"weight": 0})
                for node, data in G.nodes.data()
                if data["unit"].unit_type == "heating"
                or data["unit"].unit_type == "hotWater"
            ]
        )

        heating_primary = Unit("supplyPoint", {"id": 0, "name": "Heat"})
        G.add_node(heating_primary.id, name=heating_primary.name, unit=heating_primary)
        G.add_edge(heating_primary.id, building.units["main"][0].id, weight=0)

    # Manually add edges to Main (electricity)
    main_electricity = [
        node
        for node, data in G.nodes.data()
        if data["unit"].name == "Main" and data["unit"].unit_type == "electricity"
    ]

    if main_electricity:
        # Add other meters (not main) to main
        G.add_edges_from(
            [
                (main_electricity[0], node, {"weight": 0})
                for node, data in G.nodes.data()
                if data["unit"].unit_type == "electricity"
                and node != main_electricity[0]
            ]
        )

        electricity_primary = Unit("supplyPoint", {"id": 1, "name": "Electricity"})
        G.add_node(
            electricity_primary.id,
            name=electricity_primary.name,
            unit=electricity_primary,
        )
        G.add_edge(electricity_primary.id, main_electricity[0], weight=0)

    return G


def obtain_loop(graph, id):
    """
    Function which first finds a tree, based on id of origin unit/node.
    The end nodes of the tree are then connected back to the origin, to
    form a loop/circuit -- for circuit analysis.
    """

    # First traverse and find tree based on starting point
    tree = nx.dfs_tree(graph, id)

    # Find end nodes
    end_nodes = [n for n in tree.nodes() if tree.out_degree(n) == 0]

    # Subgraph (copy, because we want to manipulate with it)
    Gi = graph.subgraph(tree)
    Gi = Gi.copy()

    # Close loop to initial node to form circuit
    Gi.add_edges_from([(n, id, {"weight": 0}) for n in end_nodes])

    # Return tree and graph
    return tree, Gi


def load_data_graph(graph, start_date, end_date, time_resolution):
    """
    Loads data for a graph. Loops through nodes (units) and loads
    data on the units with the given start, end and resolution.
    Does not return anything -- data is held in the Unit objects.
    """
    for node, data in graph.nodes.data():
        data["unit"].load_data(start_date, end_date, time_resolution)


def draw_graph(graph):
    """
    Draws graph using networkx built in draw function.
    This is not pretty, but will help to see if graph is constructed correctly.
    """
    nx.draw(
        graph,
        pos=nx.shell_layout(graph),
        node_size=20,
        labels={node: data["unit"].name for node, data in graph.nodes.data()},
        font_size=8,
    )



import warnings
from datetime import timezone

import numpy as np
import pandas as pd
from dateutil.tz import gettz

from .api import api_get, api_string_to_datetime
from .helpers import list_to_string, timestep_start

utc = timezone.utc


class SupplyPoint:
    def __init__(self, data):
        self.id = data["id"]
        self.name = data["name"]
        self.sub_type = data["subType"]
        self.type = data["type"]
        self.__unit_ids = data["unitIds"]
        self.__applied_price_components = None
        self.__price_data = None

    def get_price_components(self):
        if self.__applied_price_components is None:
            self.__load_price_components()
        return self.__applied_price_components

    def __load_price_components(self):
        resp = api_get(
            "supplypoints/pricecomponents", out="json", payload={"ids": str(self.id)}
        )
        self.__applied_price_components = []
        for pc_i in resp[f"{self.id}"]:
            self.__applied_price_components.append(AppliedPriceComponent(pc_i))
        self.__price_data = PriceDataCollection(self.__applied_price_components)

    def get_all_price_component_ids(self):
        return [pc_i.price_component.id for pc_i in self.get_price_components()]

    def load_price_data(self, start_time, end_time):
        url_postfix = "pricecomponents/prices"
        payload = {
            "ids": list_to_string(self.get_all_price_component_ids()).replace(" ", ""),
            "start_time": start_time.isoformat(),
            "end_time": end_time.isoformat(),
            "timestamp_type": "iso_8601",
        }
        res = api_get(url_postfix, out="json", payload=payload)
        self.__price_data.load_data(res, start=start_time, end=end_time)

    def clear_price_data(self):
        return self.__price_data.clear_data()

    def get_price_data(self):
        if self.__price_data is None:
            self.get_price_components()
        return (
            self.__price_data.get_data(),
            self.__price_data.price_components_details(),
        )


class PriceDataCollection:
    def __init__(self, applied_price_components):
        self.__applied_price_components = applied_price_components
        self.__data = None

    def load_data(self, price_component_data, start, end):
        tz_out = utc
        start_hour = timestep_start("hour", t=start).astimezone(tz_out)
        end_hour = timestep_start("hour", t=end).astimezone(tz_out)
        df_out = pd.DataFrame(
            index=pd.date_range(start_hour, end_hour, freq="H", tz=tz_out)
        )

        for k_i in price_component_data.keys():
            v_i = int(k_i)
            fill_method = None
            pc_ki = price_component_data[k_i]

            # Structuring price data (either fixed, timeseries, or None) to a unique format
            if (pc_ki["fixedPriceData"] is not None) and (
                len(pc_ki["fixedPriceData"]) > 0
            ):
                pc_i = self.__get_price_component(id=v_i)
                tz_local_i = pc_i.get_timezone()

                df_raw_i = pd.DataFrame(pc_ki["fixedPriceData"])
                positions_i = []
                prices_i = []
                for j in df_raw_i["data"].index:
                    df_j = pd.DataFrame(df_raw_i["data"][j])
                    positions_i.append(df_j["position"])
                    prices_i.append(df_j["price"])
                df_raw_i["position"] = positions_i
                df_raw_i["price"] = prices_i
                if "data" in df_raw_i.keys():
                    df_raw_i.drop(columns=["data"], inplace=True)
                for j in ["validFrom", "validTo"]:
                    df_raw_i[j] = pd.to_datetime(df_raw_i[j])

                df_price_raw_i = None
                for index_ij, row_ij in df_raw_i.iterrows():
                    if row_ij["validTo"] >= start:
                        if row_ij["resolution"] == "PT1H":
                            df_base_ij = pd.DataFrame(
                                index=pd.date_range(
                                    row_ij["validFrom"],
                                    row_ij["validTo"],
                                    freq="H",
                                    closed="left",
                                )
                            )
                            df_base_ij["hour"] = df_base_ij.index.tz_convert(
                                tz_local_i
                            ).hour
                            df_price_ij = pd.DataFrame(
                                {
                                    "hour": row_ij["position"] - 1,
                                    "price": row_ij["price"],
                                }
                            )
                            df_i = df_base_ij.merge(
                                df_price_ij, how="left", left_on="hour", right_on="hour"
                            )
                            df_i.set_index(df_base_ij.index, inplace=True)
                        else:
                            if row_ij["resolution"] == "P1D":
                                freq2use = "D"
                            elif row_ij["resolution"] == "P1M":
                                freq2use = "MS"
                            else:
                                raise Exception(
                                    f"Unsupported time resolution {row_ij['resolution']}"
                                )

                            df_i = pd.DataFrame(
                                index=pd.date_range(
                                    row_ij["validFrom"],
                                    row_ij["validTo"],
                                    freq=freq2use,
                                    closed="left",
                                )
                            )

                            dt_hours = (
                                pd.Timestamp(row_ij["validTo"])
                                - pd.Timestamp(row_ij["validFrom"])
                            ).total_seconds() / 3600
                            fill_method = "ffill"
                            df_i["price"] = row_ij["price"][0] / dt_hours

                        if df_price_raw_i is None:
                            df_price_raw_i = df_i.copy()
                        else:
                            df_price_raw_i = df_price_raw_i.append(df_i.copy())

                df_price_i = df_price_raw_i[["price"]]

            elif (pc_ki["timeSeriesData"] is not None) and (
                len(pc_ki["timeSeriesData"]) > 0
            ):
                df_price_i = pd.DataFrame(pc_ki["timeSeriesData"])
                if len(df_price_i) > 0:
                    df_price_i.set_index(
                        pd.to_datetime(df_price_i["time"], infer_datetime_format=True),
                        inplace=True,
                    )
                    df_price_i.drop(columns=["time"], inplace=True)
                else:
                    df_price_i = None

            else:
                df_price_i = None

            if (df_price_i is not None) and ("price" in df_price_i.keys()):
                df_price_i = df_price_i.reindex(df_out.index, method=fill_method)
                df_out[v_i] = df_price_i["price"]
            else:
                df_out[v_i] = np.nan

        self.__data = df_out

    def clear_data(self):
        self.__data = None

    def get_data(self, application_context=None, unit_component_type=None):
        details = self.price_components_details(
            application_context=application_context,
            unit_component_type=unit_component_type,
        )

        name_dict = {}
        for index, row in details.iterrows():
            name_dict[index] = row["name"]

        return self.__data[name_dict.keys()].rename(mapper=name_dict)

    def price_components_details(
        self, application_context=None, unit_component_type=None
    ):
        df_out = pd.DataFrame(
            [
                {
                    "id": int(apc_i.price_component.id),
                    "name": apc_i.price_component.name,
                    "description": apc_i.price_component.description,
                    "unit": apc_i.price_component.unit,
                    "application_context": apc_i.application_context,
                    "billing_context": apc_i.billing_context,
                    "unit_component_type": apc_i.price_component.unit_component_type,
                    "type": apc_i.price_component.type,
                }
                for apc_i in self.__applied_price_components
            ]
        )

        # Refining results if needed
        if application_context is not None:
            df_out = df_out.loc[df_out["application_context"] == application_context, :]
        if unit_component_type is not None:
            df_out = df_out.loc[df_out["unit_component_type"] == unit_component_type, :]

        return df_out.set_index("id")

    def __get_price_component(self, id=None, name=None, price_type=None):
        if id is not None:
            f_val = lambda x: (x.price_component.id == id)
            criterion_value = ("id", id)
        elif name is not None:
            f_val = lambda x: (x.price_component.name == name)
            criterion_value = ("name", name)
        elif price_type is not None:
            f_val = lambda x: (x.price_component.type == price_type)
            criterion_value = ("type", price_type)
        else:
            raise Exception(
                "One of the following parameters must be provided: id, name"
            )

        for apc_i in self.__applied_price_components:
            if f_val(apc_i):
                return apc_i.price_component

        raise Exception(
            f"Price component not found ({criterion_value[0]}={criterion_value[1]})"
        )

    def compute_price_from_demand(self, demand_data, unit="DKK"):
        warnings.warn(
            "Usage of this function is not recommended as it is still in prototyping stage"
        )
        details = self.price_components_details()


class AppliedPriceComponent:
    def __init__(self, data):
        self.id = data["id"]
        self.supplyPointId = data["supplyPointId"]
        self.price_component = PriceComponent(data["priceComponent"])
        self.contractId = data["contractId"]
        self.created = api_string_to_datetime(data["created"])
        if data.get("validFrom") is None:
            self.validFrom = None
        else:
            self.validFrom = api_string_to_datetime(data["validFrom"])

        if data.get("validTo") is None:
            self.validTo = None
        else:
            self.validTo = api_string_to_datetime(data["validTo"])

        self.billing_context = data["billingContext"]
        self.application_context = data["applicationContext"]


class PriceComponent:
    def __init__(self, data):
        self.id = data["id"]
        self.name = data["name"]
        self.description = data["description"]
        self.created = api_string_to_datetime(data["created"])
        if data["type"] is None:
            self.type = None
            self.type_id = None
        else:
            self.type = data["type"]["Name"]
            self.type_id = data["type"]["id"]
        self.unit = data["unit"]
        self.format_type = data["formatType"]
        self.application_type = data["applicationType"]
        self.unit_component_type = data["unitComponentType"]
        self.timezone = data["timezone"]

        if data["priceArea"] is not None:
            self.price_area = PriceArea(data["priceArea"])
        else:
            self.price_area = None
        if data["authority"] is not None:
            self.authority = Authority(data["authority"])
        else:
            self.authority = None

    def get_timezone(self):
        return gettz(self.timezone)


class Authority:
    def __init__(self, data):
        self.id = data["id"]
        self.name = data["name"]


class PriceArea:
    def __init__(self, data):
        self.id = data["id"]
        self.name = data["name"]
        self.supply_type = data["supplyType"]



from preheat_open import test


class TestBuilding(test.PreheatTest):
    def test_zone_type(self, building_with_data):
        apt_zone = building_with_data.get_zones(zone_ids=[4832])[0]
        wet_zone = building_with_data.get_zones(zone_ids=[4833])[0]
        dry_zone = building_with_data.get_zones(zone_ids=[4835])[0]

        type_1, wet_or_dry_1 = apt_zone.get_type()
        assert type_1 == "apartment"
        assert wet_or_dry_1 == "?"

        type_a, wet_or_dry_a = wet_zone.get_type()
        assert type_a == "bathroom"
        assert wet_or_dry_a == "wet"

        type_b, wet_or_dry_b = dry_zone.get_type()
        assert type_b == "corridor"
        assert wet_or_dry_b == "dry"

    def test_get_units(self, building_with_data):
        # Zone A1 -- connected to unit named 'indoor_climate_1'
        zone_A1 = building_with_data.get_zones(zone_ids=[4830])[0]
        units_A1 = zone_A1.get_units("indoorClimate")
        assert units_A1[0] == building_with_data.qu(name="indoor_climate_1")
        zone_A1b = building_with_data.get_zone(id=4830)
        assert zone_A1 == zone_A1b

        # Zone A2 -- not connected to any units
        zone_A2 = building_with_data.get_zones(zone_ids=[4831])[0]
        units_A2 = zone_A2.get_units("indoorClimate")
        assert not units_A2

        zone_A = zone_A1b.get_parent_zone()
        assert zone_A.get_parent_zone() is None
        assert len(zone_A.get_units("indoorClimate")) > 0



class Component(object):
    """Defines a Component in the PreHEAT sense"""

    def __init__(self, component_data):
        # Identifier of the component
        self.cid = component_data.pop("cid", None)
        self.id = component_data.pop("id", None)

        # Name of the component
        self.name = component_data.pop("name", str(self.id))

        # Tag (e.g. BACNET/source name)
        self.tag = component_data.pop("tag", None)

        # Data for the component (PreHEAT_API.Data)
        self.data = None

        # Factor to divide by to obtain a 'standard' unit
        self.std_unit_devisor = component_data.pop("stdUnitDivisor", 1)

        # Standard unit for this type of component
        self.std_unit = component_data.pop("stdUnit", "")

    def __repr__(self):
        return "{0}({1}, {2})".format(type(self).__name__, self.name, self.tag)



import pandas as pd
import pytest

from preheat_open import test


class TestBuilding(test.PreheatTest):
    def test_load_data(self, unit_with_data, short_period):
        assert unit_with_data is not None
        assert unit_with_data.data is not None
        assert not unit_with_data.data.empty
        assert unit_with_data.has_data()
        print(unit_with_data.data)
        assert unit_with_data.has_data("supplyT")
        with pytest.warns(RuntimeWarning):
            unit_with_data.load_data(*short_period)
            assert not unit_with_data.data.empty

    def test_clear_data(self, unit_with_data):
        unit_with_data.clear_data()
        assert unit_with_data.data is not None
        assert unit_with_data.data.empty

    def test_selected_data_loading(self, unit, short_period):
        unit.load_data(*short_period, components=["energy", "power"])
        assert unit.data is not None
        assert not unit.data.empty
        assert "energy" in unit.data.keys()
        assert "power" in unit.data.keys()
        assert "volume" not in unit.data.keys()
        assert "supplyT" not in unit.data.keys()

    def test_get_state(self, unit, short_period):
        s = unit.get_state(update=True, estimate="last", seconds_back=300)
        # The location does not get new data
        assert len(s) == 0
        # retrieve old state
        t_now = short_period[0]
        s = unit.get_state(update=True, estimate="last", seconds_back=300, t_now=t_now)
        assert len(s) > 0
        unit.clear_state()



import warnings
from copy import deepcopy

import pytest

import preheat_open
from preheat_open import running_in_test_mode

# Setting up a test API key, which is only valid for a dummy test installation
API_KEY = "KVkIdWLKac5XFLCs2loKb7GUitkTL4uJXoSyUFIZkVgWuCk8Uj"
ANONYMISED_API_KEY = "3xa0SeGXa4WlkrB68qGR9NoDAzVvGdiG3XAabKu6n7n5qQTDkL"
TEST_LOCATION_ID = 2756

# Warning the user that this module is not meant to be used for non test-related activities
if running_in_test_mode() is False:
    warnings.warn(
        """

The module 'preheat_open.test' is not meant to be imported and actively used, 
unless you are specifically carrying out a test.

    """
    )

SHORT_TEST_PERIOD = ("2021-05-01 00:00", "2021-05-02 00:00", "hour")


class PreheatTest:
    @pytest.fixture(autouse=True)
    def set_api_key(self):
        preheat_open.api.set_api_key(API_KEY)

    @pytest.fixture()
    def bypass_api_key(self):
        preheat_open.api.set_api_key(None)
        yield None
        preheat_open.api.set_api_key(API_KEY)

    @pytest.fixture(scope="session")
    def building_id(self):
        return TEST_LOCATION_ID

    @pytest.fixture(scope="session")
    def unit_id(self):
        return 15312

    @pytest.fixture(scope="session")
    def control_unit_id(self):
        return 15357

    @pytest.fixture(scope="session")
    def building(self, building_id):
        return preheat_open.Building(building_id)

    @pytest.fixture(scope="session")
    def building_with_data(self, building, medium_period):
        building_new = deepcopy(building)
        building_new.load_data(*medium_period)
        return building_new

    @pytest.fixture(scope="session")
    def unit(self, building, unit_id):
        return building.query_units(unit_id=unit_id)[0]

    @pytest.fixture(scope="session")
    def unit_with_data(self, unit, medium_period):
        unit_new = deepcopy(unit)
        unit_new.load_data(*medium_period)
        return unit_new

    @pytest.fixture(scope="session")
    def control_unit(self, building):
        return building.qu("control", "control_unit_custom_1")

    @pytest.fixture(scope="session")
    def weather_unit(self, building):
        return building.weather

    @pytest.fixture(scope="session")
    def short_period(self):
        start_date = SHORT_TEST_PERIOD[0]
        end_date = SHORT_TEST_PERIOD[1]
        time_resolution = "hour"
        return start_date, end_date, time_resolution

    @pytest.fixture(scope="session")
    def medium_period(self):
        start_date = "2021-05-01 00:00"
        end_date = "2021-05-07 00:00"
        time_resolution = "hour"
        return start_date, end_date, time_resolution



"""
This module defines the basics of the interaction with Neogrid's web API
"""
import json
import os
from datetime import datetime
from io import StringIO
from typing import Dict, Optional

import numpy as np
import pandas as pd
from dateutil.tz import gettz
from requests import Session
from requests.models import Response

import preheat_open

from .logging import Logging
from .singleton import Singleton

BASE_URL = "https://api.neogrid.dk/public/api/v1"
MAX_CIDS_PER_REQ = 3
MAX_POINTS_PER_REQ = 90000

# Perhaps we can steal this from the OS?
TIMEZONE = gettz("Europe/Copenhagen")
DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"

USER_LEVEL_CONFIGURATION_FILE = os.path.expanduser("~/.preheat/config.json")
MACHINE_LEVEL_CONFIGURATION_FILE = os.path.expanduser("/etc/preheat/config.json")
MACHINE_LEVEL_CONFIGURATION_FILE_OPT = os.path.expanduser(
    "/etc/opt/preheat/config.json"
)


def configuration_file_path() -> str:
    """

    :return:
    :rtype:
    """
    if os.path.exists(USER_LEVEL_CONFIGURATION_FILE):
        out = USER_LEVEL_CONFIGURATION_FILE
    elif os.path.exists(MACHINE_LEVEL_CONFIGURATION_FILE_OPT):
        out = MACHINE_LEVEL_CONFIGURATION_FILE_OPT
    elif os.path.exists(MACHINE_LEVEL_CONFIGURATION_FILE):
        out = MACHINE_LEVEL_CONFIGURATION_FILE
    else:
        raise MissingConfigurationFile()
    return out


class APIKeyMissingError(Exception):
    pass


class AccessDeniedError(Exception):
    pass


class APIDataExtractError(Exception):
    pass


class MissingConfigurationFile(Exception):
    pass


def api_string_to_datetime(t: str) -> datetime:
    """

    :param t:
    :type t:
    :return:
    :rtype:
    """
    return datetime.strptime(t, "%Y-%m-%dT%H:%M:%S.%f%z")


def datetime_to_api_string(t: datetime) -> str:
    """

    :param t:
    :type t:
    :return:
    :rtype:
    """
    return t.isoformat()


def ids_list_to_string(list2use, separator: str = ",") -> str:
    """Helper function to turn list into string, e.g. comma separated (default).

    :param list2use:
    :type list2use:
    :param separator:
    :type separator:
    :return:
    :rtype:

    """

    if isinstance(list2use, list):
        res = separator.join(map(str, list2use))
    elif isinstance(list2use, pd.Series):
        res = separator.join(map(str, list2use.to_list()))
    elif isinstance(list2use, np.ndarray):
        res = separator.join(map(str, list2use.tolist()))
    else:
        res = str(list2use)

    return res


def load_configuration() -> Dict:
    """

    :return:
    :rtype:
    """
    config = None
    try:
        # Prefer config file in user home dir, else use the global config file
        config_file = configuration_file_path()
        if os.path.exists(config_file):
            Logging().info(f"Loading config from {config_file}")
            with open(config_file, "r") as fp:
                config = json.load(fp)
    except:
        config = None

    return config


class ApiSession(metaclass=Singleton):
    """Classs for something about Apisessions???"""

    def __init__(self):
        self.base_url = None
        self.session = Session()
        Logging().debug(f"Creating API session [{self.addr()}]...")

        self.set_api_url()
        self.set_api_key()

    def addr(self):
        """

        :return:
        :rtype:
        """
        return hex(id(self.session))

    def get_api_key(self) -> str:
        """

        :return:
        :rtype:
        """

        return self.session.headers.get("Authorization")

    def set_api_key(self, api_key: Optional[str] = None) -> None:
        """

        :param api_key:
        :type api_key:
        :return:
        :rtype:
        """
        if api_key is None:
            api_key = self._load_api_key(
                self.get_api_key().split(" ")[1]
                if self.get_api_key() is not None
                else None
            )
        else:
            Logging().debug(f"API key set manually [{self.addr()}]")
        headers = {"Authorization": "Apikey " + str(api_key)}
        self.session.headers.update(headers)

    def set_api_url(self, api_url=None):
        """

        :param api_url:
        :type api_url:
        :return:
        :rtype:
        """
        if api_url is None:
            api_url = self._load_api_url()
        else:
            Logging().debug(f"API url set manually [{self.addr()}]")
        self.base_url = api_url

    def api_get(self, endpoint, out="json", payload=None, headers=None):
        """

        :param endpoint:
        :type endpoint:
        :param out:
        :type out:
        :param payload:
        :type payload:
        :param headers:
        :type headers:
        :return:
        :rtype:
        """
        payload = payload if payload is not None else {}
        headers = headers if headers is not None else {}

        Logging().info(f"GET /{endpoint} [{self.addr()}]")

        path = f"{self.base_url}/{endpoint}"

        if out == "csv":
            headers["Accept"] = "text/csv"
        resp = self.session.get(path, params=payload, headers=headers)
        resp.raise_for_status()

        try:
            if out == "json":
                return resp.json()
            elif out == "csv":
                data = StringIO(resp.text)
                return pd.read_csv(data, delimiter=";")
        except:
            raise APIDataExtractError()

    def api_put(self, endpoint, json_payload=None) -> Response:
        """

        :param endpoint:
        :type endpoint:
        :param json_payload:
        :type json_payload:
        :return:
        :rtype:
        """
        json_payload = json_payload if json_payload is not None else {}

        Logging().info(f"PUT /{endpoint} [{self.addr()}]")

        path = f"{self.base_url}/{endpoint}"
        r = self.session.put(path, json=json_payload)
        r.raise_for_status()
        return r

    def _load_api_key(self, existing):
        """Load API key from environment variable
        - if API key not present in environment variable it will be set None
        - will then break when trying to access data on a building

        :param existing:
        :type existing:
        :return:
        :rtype:

        """

        # First respect any override by set_api_key
        if os.environ.get("PREHEAT_API_KEY_OVERRIDE") is not None:
            if (
                existing is None
                or os.environ.get("PREHEAT_API_KEY_OVERRIDE") != existing
            ):
                Logging().info(f"Using API key from set_api_key() [{self.addr()}]")
            return os.environ.get("PREHEAT_API_KEY_OVERRIDE")

        # Else, try PREHEAT_API_KEY environment variable
        elif os.environ.get("PREHEAT_API_KEY") is not None:
            if existing is None or os.environ.get("PREHEAT_API_KEY") != existing:
                Logging().info(
                    f"Using API key from environment variable [{self.addr()}]"
                )
            return os.environ.get("PREHEAT_API_KEY")

        # Else, try loading configuration file
        else:
            # If config loaded, set API key accordingly to file
            if preheat_open.config is not None:
                config_keys = preheat_open.config.keys()
                if "API_KEY" in config_keys:
                    if existing is None or preheat_open.config["API_KEY"] != existing:
                        Logging().info(
                            f"Using API key from config file [{self.addr()}]"
                        )
                    return preheat_open.config["API_KEY"]
                elif "PREHEAT_API_KEY" in config_keys:
                    if (
                        existing is None
                        or preheat_open.config["PREHEAT_API_KEY"] != existing
                    ):
                        Logging().info(
                            f"Using API key from config file [{self.addr()}]"
                        )
                    return preheat_open.config["PREHEAT_API_KEY"]
                else:
                    raise Exception(
                        f"""Your configuration file is missing a 'PREHEAT_API_KEY' field"""
                    )

            # No environment variable, no config file, nothing we can do
            else:
                return str(None)

    def _load_api_url(self):
        """

        :return:
        :rtype:
        """
        if os.environ.get("PREHEAT_API_URL_OVERRIDE") is not None:
            url = os.environ.get("PREHEAT_API_URL_OVERRIDE")
            Logging().info(f"Using API url: {url} [{self.addr()}]")
            return url
        else:
            Logging().info(f"Using default API url [{self.addr()}]")
            return BASE_URL


def set_api_key(api_key: str) -> None:
    """

    :param api_key:
    :type api_key:
    :return:
    :rtype:
    """
    if api_key is None:
        if "PREHEAT_API_KEY_OVERRIDE" in os.environ:
            os.environ.pop("PREHEAT_API_KEY_OVERRIDE")
            Logging().info("API key override unset")
    else:
        os.environ["PREHEAT_API_KEY_OVERRIDE"] = api_key
        Logging().info("API key override set")
    ApiSession().set_api_key()


def set_api_url(api_url: str) -> None:
    """

    :param api_url:
    :type api_url:
    :return:
    :rtype:
    """
    if api_url is None:
        if "PREHEAT_API_URL_OVERRIDE" in os.environ:
            os.environ.pop("PREHEAT_API_URL_OVERRIDE")
            Logging().debug("API key override unset")
    else:
        os.environ["PREHEAT_API_URL_OVERRIDE"] = api_url
        Logging().debug("API key override set")
    ApiSession().set_api_url()


def api_get(endpoint: str, out: str = "json", payload=None, headers=None) -> Response:
    """

    :param endpoint:
    :type endpoint:
    :param out:
    :type out:
    :param payload:
    :type payload:
    :param headers:
    :type headers:
    :return:
    :rtype:
    """
    return ApiSession().api_get(endpoint, out, payload, headers)


def api_put(endpoint: str, json_payload=None) -> Response:
    """

    :param endpoint:
    :type endpoint:
    :param json_payload:
    :type json_payload:
    :return:
    :rtype:
    """
    return ApiSession().api_put(endpoint, json_payload)



from __future__ import annotations

from typing import List, Tuple

from .building_unit import BaseBuildingUnit


def populate_zones(zones_data, parent_zone=None):
    zones = []
    # Loop over zones; create and populate Zone instances
    for zone_i in zones_data:
        zones.append(Zone(zone_i, parent_zone=parent_zone))

    return zones


class Zone(object):
    """Defines a building zone in the PreHEAT sense"""

    def __init__(self, zone_data, parent_zone=None):
        # Identifier of the zone
        self.id = zone_data.pop("id")
        # Name of the zone
        self.name = zone_data.pop("name", str(self.id))
        # Floor area of the zone
        self.area = zone_data.pop("zoneArea", None)
        # Whether zone has an external wall (boolean)
        self.has_external_wall = zone_data.pop("hasExternalWall", None)
        # Sub-zones of the zone (list of PreHEAT_API.Zones)
        self.sub_zones = populate_zones(zone_data.pop("subZones", []), parent_zone=self)
        # Adjacency
        self.adjacent_zones = [
            z["zoneId"] for z in zone_data.pop("adjacentZones", None)
        ]
        # Type of the zone
        self._type = zone_data.pop("type")

        # Parent zone
        self.__parent = parent_zone

        # Adding coupled units
        self.__coupled_units = dict()

    def get_sub_zones(self, zone_ids=None) -> List[Zone]:
        if zone_ids is None:
            res = self.sub_zones
        else:
            res = []
            for z_i in self.sub_zones:
                if z_i.id in zone_ids:
                    res.append(z_i)
                res += z_i.get_sub_zones(zone_ids=zone_ids)
        return res

    def get_parent_zone(self) -> Zone:
        return self.__parent

    def get_type(self) -> Tuple[str, str]:
        """
        Returns the type of the zone ('room', 'stairway', 'corridor', 'bathroom', 'kitchen')
        and its 'dryness' ('dry', 'wet', '?')

        :return: zone_type, wet_or_dry
        """
        zone_type = self._type.lower()
        if zone_type in ["room", "stairway", "corridor"]:
            wet_or_dry = "dry"
        elif zone_type in ["bathroom", "kitchen"]:
            wet_or_dry = "wet"
        else:
            wet_or_dry = "?"
        return zone_type, wet_or_dry

    def __repr__(self):
        return "{0}({1})".format(type(self).__name__, self.name)

    def add_coupled_unit(self, unit) -> None:
        unit_type = unit.unit_type
        if unit_type in self.__coupled_units.keys():
            if unit.id in [u_i.id for u_i in self.__coupled_units[unit_type]]:
                pass  # Skip addition of already existing units
            else:
                self.__coupled_units[unit_type].append(unit)
        else:
            self.__coupled_units[unit_type] = [unit]

    def get_units(self, unit_type, sub_zones=True) -> List[BaseBuildingUnit]:
        """
        Returns the coupled units of a given type

        :param unit_type: str
        :param sub_zones: bool (True)
        :return: list(BuildingUnit) or None
        """
        us = self.__coupled_units.get(unit_type)
        if sub_zones is True:
            for z_i in self.sub_zones:
                us_i = z_i.get_units(unit_type, sub_zones=True)
                if us_i is not None:
                    if us is None:
                        us = us_i
                    else:
                        us += us_i
        return us



import setuptools

import versioneer

documentation_requirements = [
    "Sphinx",
    "m2r2",
    "sphinx-autodoc-typehints",
    "sphinx-rtd-theme",
    "nbsphinx",
    "pandoc",
]

setuptools.setup(
    name="preheat_open",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="Neogrid and contributors",
    author_email="analytics@neogrid.dk",
    description="Python wrapper for Neogrid Technologies' REST API",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://gitlab.com/neogrid-technologies-public/preheat-open-python",
    project_urls={
        "Bug Tracker": "https://gitlab.com/neogrid-technologies-public/preheat-open-python/-/issues",
        "Documentation": "https://neogrid-technologies-public.gitlab.io/preheat-open-python",
        "Source Code": "https://gitlab.com/neogrid-technologies-public/preheat-open-python",
    },
    packages=setuptools.find_packages(),
    install_requires=["dateutils", "requests", "pandas", "numpy", "networkx"],
    extras_require={
        "doc": documentation_requirements,
        "dev": [
            "setuptools>=42",
            "wheel",
            "pytest",
            "pytest-cov",
            "pytest-xdist",
        ]
        + documentation_requirements,
    },
)



import pandas as pd
import pytest

from preheat_open import test


class TestBuilding(test.PreheatTest):
    def test_load_data(self, weather_unit, short_period):
        weather_unit.load_data(*short_period)
        assert weather_unit is not None
        assert weather_unit.data is not None
        assert not weather_unit.data.empty
        assert weather_unit.has_data()
        print(weather_unit.data)
        assert weather_unit.has_data("Temperature")
        with pytest.warns(RuntimeWarning):
            weather_unit.load_data(*short_period)
            assert not weather_unit.data.empty

    def test_clear_data(self, weather_unit, short_period):
        weather_unit.load_data(*short_period)
        weather_unit.clear_data()
        assert weather_unit.data is not None
        assert weather_unit.data.empty

    def test_selected_data_loading(self, weather_unit, unit, short_period):
        weather_unit.load_data(*short_period, components=["Temperature", "LowClouds"])
        assert weather_unit.data is not None
        assert not weather_unit.data.empty
        assert "Temperature" in weather_unit.data.keys()
        assert "LowClouds" in weather_unit.data.keys()
        assert "Humidity" not in weather_unit.data.keys()
        assert "Fog" not in weather_unit.data.keys()

    def test_get_state(self, weather_unit):

        sw1 = weather_unit.get_state(
            update=True, estimate="last", seconds_back=12 * 3600
        )
        sw2 = weather_unit.get_state(update=False, estimate="median")
        sw3 = weather_unit.get_state(update=False, estimate="mean")
        assert isinstance(sw1, pd.Series)
        assert isinstance(sw2, pd.Series)
        assert isinstance(sw3, pd.Series)

        # The location does not get new data
        assert len(sw1) > 0
        assert len(sw2) > 0
        assert len(sw3) > 0
        weather_unit.clear_state()
